#include "..\..\common\ImageFilter.h"
#include <string.h>

#pragma warning (disable:4800)

using namespace FilterSpace;
using namespace System::Windows::Forms;


void FreeUpHistogramMemory(int** hist, int num)
{
	for (int i = 0; i < num; i++) delete [] hist[i];
	delete [] hist;
}


int calculate_threshold(int* hist, int image_size)
{
	double l_var, r_var, variances[257];
	int    d, i, j, l_mean[256], r_mean[256], l_sum[256], r_sum[256], min = 0;

	l_mean[0]   = 0; 
	l_sum[0]    = hist[0];
	r_mean[255] = 255 * hist[255];
	r_sum[255]  = hist[255];

	// tlagok elszmtsa
	for (j = 1; j < 256; ++j)
	{
		l_mean[j]       = l_mean[j-1]     + hist[j] * j;               // A prefix vrhat rtke nem leosztva a pixelszmmal
		l_sum[j]        = l_sum[ j-1]     + hist[j];                   // A prefixbe es pixelek szma
		r_mean[255 - j] = r_mean[256 - j] + hist[255 - j] * (255 - j); // A prefix vrhat rtke nem leosztva a pixelszmmal
		r_sum[255 - j]  = r_sum[ 256 - j] + hist[255 - j];             // A prefixbe es pixelek szma
	}

	// prefix tlagok
	for (j = 0; j < 256; ++j)
	{
		if (l_sum[j]) l_mean[j] /= l_sum[j];
		if (r_sum[j]) r_mean[j] /= r_sum[j];
	}

	for (j = 0; j < 256; ++j)
	{
		r_var = l_var = 0;

		// LEFT VARIANCE
		for (i = 0; i < j; ++i)
		{
			d = i - l_mean[j];
			l_var += d * d * hist[i];
		}

		// RIGHT VARIANCE
		for (i = j; i < 256; ++i)
		{
			d = i - r_mean[j];
			r_var += d * d * hist[i];
		}

		variances[j] = (l_var + r_var) / (image_size);
	}

	for (j = 1; j < 256; ++j) if ( variances[j] < variances[min] ) min = j;

	return min;
} 






/////////////////////////////
//       FILTER-PASS1      //
/////////////////////////////

bool FilterCorePass1(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	int width  = (sw) ? sw : f->width;
	int height = (sh) ? sh : f->height;
	int x, y, processed_pixels = 0;
	int* curr_block_hist = f->block_histogram[f->nCoreThreadId];

	for (y = 0; y < height; y++)
	{
		unsigned char* p = f->datain + (y + sy) * f->stridein + sx;
		for (x = 0; x < width; x++) curr_block_hist[*(p++)]++;

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;

		// CONDITION CAN BE COMBINED WITH E.G. " && (processed_pixels > 1000)" 
		// TO AVOID FREQUENT MUTEX LOCKS

		if (updater != nullptr && processed_pixels > 5000)
		{
			if (updater->WaitOne(10, false))
			{
				*(f->processing_ptr) += processed_pixels;
				processed_pixels = 0;
				updater->Release(1);
			}
		}
		else
		{
			if (f->processing_ptr) *(f->processing_ptr) += processed_pixels;
			processed_pixels = 0;
		}
		
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	// FILTER CODE END -->
	if (f->processing_ptr && f->processing_size) *(f->processing_ptr) = *(f->processing_size);

	return true;
}




/////////////////////////////
//       FILTER-PASS2      //
/////////////////////////////

bool FilterCorePass2(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	int width  = (sw) ? sw : f->width;
	int height = (sh) ? sh : f->height;
	int thres  = f->threshold;
	int x, y, processed_pixels = 0;

	for (y = 0; y < height; y++)
	{
		unsigned char* p_in  = f->datain  + (y + sy) * f->stridein + sx;
		unsigned char* p_out = f->dataout + y * f->strideout;
		for (x = 0; x < width; x++) *(p_out++) = (*(p_in++) < thres) ? 0 : 255;

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;

		// CONDITION CAN BE COMBINED WITH E.G. " && (processed_pixels > 1000)" 
		// TO AVOID FREQUENT MUTEX LOCKS

		if (updater != nullptr && processed_pixels > 5000)
		{
			if (updater->WaitOne(10, false))
			{
				*(f->processing_ptr) += processed_pixels;
				processed_pixels = 0;
				updater->Release(1);
			}
		}
		else
		{
			if (f->processing_ptr) *(f->processing_ptr) += processed_pixels;
			processed_pixels = 0;
		}
		
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	// FILTER CODE END -->
	if (f->processing_ptr && f->processing_size) *(f->processing_ptr) = *(f->processing_size);

	return true;
}




IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, stride = 0, height = 0, run_local = 1;

	if (!fs) return IntPtr(0);

	xml_node^ xn = gcnew xml_node( fs );

	filtersettings* f = XmlToMyStruct( xn, true );

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* fb = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(fb);
	delete [] fb;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			xt->XmlGetChildValueByName("Stride", stride);
			xt->XmlGetChildValueByName("Height", height);
			if (stride * height == 0) return IntPtr(0);
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", outdata);
		}
	}
	xn->children.Add( xt );

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}



filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	int packed, interleaved;
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain))
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) )  ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) )   ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;

	return filset;
}


bool ImageFilter::CanLocalize(){ return true; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,           1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}






// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return false; // (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr) = 0;
		*(filset->processing_size) = filset->width * filset->height;

		int cpucores = System::Environment::ProcessorCount;

		// PASS-1: Creating histogram

		filset->block_histogram = new int*[cpucores];
		for (int i = 0; i < cpucores; i++) 
		{
			filset->block_histogram[i] = new int[256];
			memset(filset->block_histogram[i], 0, sizeof(int) * 256);
		}

		if (!filset->nCoreEnabled || cpucores == 1)
		{
			filset->nCoreThreadId = 0;
			b = ::FilterCorePass1(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
		}
		else
		{
			filter_core_func f = (::FilterCorePass1);
			filset->nCoreExecutionSize = 1; // SAFETY CONDITIONING
			b = ImageFilterBase::ExecuteFilterBlock(filset, f);
		}
		*(filset->processing_ptr) = 0;

		// CUMULATE PASS-1 RESULTS

		for (int i = 1; i < cpucores; i++) for (int j = 0; j < 256; j++) filset->block_histogram[0][j] += filset->block_histogram[i][j];
		filset->threshold = calculate_threshold(filset->block_histogram[0], *(filset->processing_size));

		// PASS-2: THRESHOLDING

		if (*(filset->nCoreKillSignal) == 0)
		{
			if (!filset->nCoreEnabled || cpucores == 1)
			{
				filset->nCoreThreadId = 0;
				b = ::FilterCorePass2(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
			}
			else
			{
				filter_core_func f = (::FilterCorePass2);
				filset->nCoreExecutionSize = 1; // SAFETY CONDITIONING
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
		}

		// FREE UP, SET RESULT VALUES, RETURN...
		FreeUpHistogramMemory(filset->block_histogram, cpucores);
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
		delete filset;
	}
	return b;
}

